@@ -40,7 +40,7 @@ function isElementInViewport(el) {
 }
 
 require.config({ paths: { 'vs': '/lib/vs' }});
-define(["vs/editor/editor.main"], (monaco) => {
+define(["vs/editor/editor.main", "Fuse"], (monaco, Fuse) => {
 
 const KM = monaco.KeyMod;
 const KC = monaco.KeyCode;
@@ -49,6 +49,86 @@ let mainRepl = null;
 let evalIdGen = 0;
 
 
+class FuzzySearchResults {
+    constructor(element, entries) {
+        this.element = element;
+        this.entries = entries;
+        this.structured_results = null;
+        this.cursors = [0];
+        element.className = "snek-fuzzy-search";
+        this.results = document.createElement("div");
+        this.results.className = "snek-fuzzy-search-results";
+        element.appendChild(this.results);
+    }
+
+    get() {
+        let ret = [];
+        for (let cursor of this.cursors) {
+            const result = this.structured_results[cursor];
+            if (result !== undefined) {
+                const entry = result.item;
+                ret.unshift(this.entries[entry]);
+            }
+        }
+        return ret.join("\n") || null;
+    }
+
+    showResults(results) {
+        this.results.innerHTML = "";
+        for (const result of results) {
+            const row = document.createElement("div");
+            row.innerText = this.entries[result.item];
+            this.results.appendChild(row);
+        }
+        if (!this.results.children.length) {
+            this.results.innerHTML = "No matches";
+        }
+        this.structured_results = results;
+    }
+
+    updateCursors(new_cursors) {
+        const n = this.structured_results.length;
+        if (n === 0) {
+            return;
+        }
+        for (let cursor of this.cursors) {
+            if (new_cursors.indexOf(cursor) === -1) {
+                this.results.children[cursor].classList.remove("snek-fuzzy-cursor");
+            }
+        }
+        for (let cursor of new_cursors) {
+            this.results.children[cursor].classList.add("snek-fuzzy-cursor");
+        }
+        this.cursors = new_cursors;
+    }
+
+    setCursor(value) {
+        let new_cursors = [value];
+        this.updateCursors(new_cursors);
+    }
+
+    addCursor(value) {
+        let new_cursors = this.cursors.slice(0);
+        new_cursors.push(value);
+        new_cursors.sort();
+        this.updateCursors(new_cursors);
+    }
+
+    expandUp() {
+        const x = this.cursors.length - 1;
+        if (this.cursors.length && this.cursors[x] < this.structured_results.length - 1) {
+            this.addCursor(this.cursors[x] + 1);
+        }
+    }
+
+    expandDown() {
+        if (this.cursors.length && this.cursors[0] > 0) {
+            this.addCursor(this.cursors[0] - 1);
+        }
+    }
+}
+
+
 class Repl {
 
     constructor(target) {
@@ -57,13 +137,20 @@ class Repl {
         this.pane = target.querySelector(".snek-pane");
         this.outerPane = target.querySelector(".snek-outer-pane");
         this.pinpane = target.querySelector(".snek-pin-pane");
+        this.inputOuter = target.querySelector(".snek-input-box");
         this.inputBox = target.querySelector(".snek-input");
         this.inputMode = target.querySelector(".snek-input-mode");
         this.statusBar = target.querySelector(".snek-status-bar");
         this.nav = target.querySelector(".snek-nav");
         this._setupEditor(this.inputBox);
-        this.historyRange = [0, 0];
+
+        this.historySelection = 0;
+        this.filter = null;
+        this.filteredHistory = [""];
         this.history = [""];
+        this.expectedContent = null;
+        this.historyPopup = null;
+
         target.onclick = this.$globalClickEvent.bind(this);
         window.onkeydown = this.$globalKDEvent.bind(this);
         window.$$SKTK = this.sktk.bind(this);
@@ -116,6 +203,8 @@ class Repl {
     }
 
     $globalClickEvent(evt) {
+        let cmdctrl = isMac ? (evt.metaKey && !evt.ctrlKey) : evt.ctrlKey;
+
         if (evt.detail === 2) {
             // Double-click
             return;
@@ -129,11 +218,11 @@ class Repl {
                 // Stop if non-default behavior or link
                 break;
             }
-            else if (evt.metaKey && target.getAttribute("pinnable") !== null) {
+            else if (cmdctrl && evt.altKey && !evt.shiftKey && target.getAttribute("pinnable") !== null) {
                 this.pin(target);
                 break;
             }
-            else if (!evt.metaKey && target.getAttribute("objid") !== null) {
+            else if (cmdctrl && !evt.altKey && !evt.shiftKey && target.getAttribute("objid") !== null) {
                 this.sktk(parseInt(target.getAttribute("objid")));
                 break;
             }
@@ -262,12 +351,15 @@ class Repl {
             folding: false,
             renderLineHighlight: "none",
             lineDecorationsWidth: 0,
+            suggestOnTriggerCharacters: false,
+            wordBasedSuggestions: false,
         });
         this.editor = editor;
         target.classList.add("snek-editor-cyclable");
         target.$editor = this.editor;
         this.editor.onDidContentSizeChange(this.event_updateHeight.bind(this));
 
+        this.popupActive = this.editor.createContextKey("popupActive", false);
         this.atBeginning = this.editor.createContextKey("atBeginning", true);
         this.atEnd = this.editor.createContextKey("atEnd", true);
         this.multiline = this.editor.createContextKey("multiline", false);
@@ -286,6 +378,14 @@ class Repl {
                 total > 1
                 || editor.getModel().getLineContent(1).match(/[\(\[\{:]$|^@/)
             )
+            if (this.historyPopup) {
+                this.setupFilter();
+            }
+            else if (this.expectedContent !== null
+                && this.editor.getValue() !== this.expectedContent) {
+                    this.filter = null;
+                    this.expectedContent = null;
+            }
         });
 
         let _submit = () => {
@@ -295,14 +395,15 @@ class Repl {
                 command: "submit",
                 expr: val,
             });
-            if (this.history[1] !== val) {
+            if (this.history[1] !== val && val !== "") {
                 this.history[0] = val;
                 this.history.unshift("");
             }
             else {
                 this.history[0] = "";
             }
-            this.historyRange = [0, 0];
+            this.filter = null;
+            this.destroyHistoryPopup();
             // The flex-direction on the outer pane is reversed, so
             // 0 scrolls it to the bottom. Handy.
             this.outerPane.scrollTop = 0;
@@ -311,7 +412,27 @@ class Repl {
         editor.addCommand(
             KC.Enter,
             _submit,
-            "!multiline"
+            "!multiline && !popupActive"
+        );
+
+        editor.addCommand(
+            KC.Enter,
+            () => {
+                const selection = this.historyPopup.get();
+                if (selection) {
+                    this.editor.setValue(selection);
+                    let nlines = this.editor.getModel().getLineCount();
+                    this.editor.setPosition({lineNumber: nlines, column: 1000000});
+                }
+                this.destroyHistoryPopup();
+            },
+            "popupActive"
+        );
+
+        editor.addCommand(
+            KC.Escape,
+            () => this.destroyHistoryPopup(),
+            "popupActive"
         );
 
         editor.addCommand(
@@ -324,15 +445,10 @@ class Repl {
             () => { this.pane.innerHTML = ""; }
         );
 
-        // editor.addCommand(
-        //     KM.Alt | KC.UpArrow,
-        //     () => { this.historyShift(-1, true); },
-        // );
-
-        // editor.addCommand(
-        //     KM.Alt | KC.DownArrow,
-        //     () => { this.historyShift(1, true); },
-        // );
+        editor.addCommand(
+            KM.WinCtrl | KC.KEY_R,
+            () => this.setupHistoryPopup()
+        );
 
         editor.addCommand(
             KC.UpArrow,
@@ -346,42 +462,99 @@ class Repl {
             "atEnd"
         );
 
+        editor.addCommand(
+            KM.Shift | KC.UpArrow,
+            () => { this.historyPopup.expandUp(); },
+            "popupActive"
+        );
+
+        editor.addCommand(
+            KM.Shift | KC.DownArrow,
+            () => { this.historyPopup.expandDown(); },
+            "popupActive"
+        );
+
         this.editor = editor;
     }
 
-    historyShift(delta, agglutinate) {
-        let [htop, hbot] = this.historyRange;
-        if (hbot === 0) {
-            this.history[0] = this.editor.getValue();
-        }
-        let n = this.history.length;
+    setupHistoryPopup() {
+        const el = document.createElement("div");
+        const fuzz = new FuzzySearchResults(
+            el,
+            this.history.slice(1),
+        );
+        this.inputOuter.insertBefore(el, this.inputMode);
+        this.historyPopup = fuzz;
+        this.popupActive.set(true);
+        this.setupFilter();
+    }
 
-        let new_htop = 0;
-        let new_hbot = 0;
-        if (delta < 0) {
-            new_htop = Math.max(0, Math.min(n - 1, htop - delta));
-            new_hbot = agglutinate ? hbot : new_htop;
+    destroyHistoryPopup() {
+        if (this.historyPopup !== null) {
+            this.historyPopup.element.remove();
+            this.historyPopup = null;
+            this.popupActive.set(false);
+            this.filter = null;
+        }
+    }
+
+    setupFilter() {
+        const popup = this.historyPopup !== null;
+        this.filter = this.editor.getValue();
+        this.history[0] = this.filter;
+        const histo = popup ? this.history.slice(1) : this.history;
+        if (!this.filter) {
+            this.filteredHistory = histo.map(
+                (elem, idx) => ({item: idx})
+            );
         }
         else {
-            new_hbot = Math.max(0, Math.min(n - 1, hbot - delta));
-            new_htop = agglutinate ? htop : new_hbot;
+            const options = {
+                includeScore: true,
+                includeMatches: true,
+            };
+            const fuse = new Fuse(histo, options);
+            this.filteredHistory = fuse.search(this.filter);
+        }
+        this.historySelection = 0;
+        if (popup) {
+            this.historyPopup.showResults(this.filteredHistory);
+            this.historyPopup.setCursor(0);
+        }
+    }
+
+    historyShift(delta) {
+        if (this.filter === null) {
+            this.setupFilter();
         }
 
-        if (new_hbot == hbot && new_htop == htop) {
+        let sel = this.historySelection;
+        let n = this.filteredHistory.length;
+        let new_sel = Math.max(0, Math.min(n - 1, sel - delta));
+
+        if (new_sel === sel) {
             return;
         }
 
-        this.historyRange = [new_htop, new_hbot];
-        let text = this.history.slice(new_hbot, new_htop + 1).reverse().join("\n");
+        this.historySelection = new_sel;
+        let item = this.filteredHistory[new_sel].item;
 
-        this.editor.setValue(text);
-
-        if (delta < 0) {
-            let nlines = this.editor.getModel().getLineCount();
-            this.editor.setPosition({lineNumber: nlines, column: 1000000});
+        if (this.historyPopup !== null) {
+            this.historyPopup.setCursor(new_sel);
         }
         else {
-            this.editor.setPosition({lineNumber: 1, column: 1000000});
+            let text = this.history[item];
+
+            this.expectedContent = text;
+            this.editor.setValue(text);
+
+            if (delta < 0) {
+                let nlines = this.editor.getModel().getLineCount();
+                this.editor.setPosition({lineNumber: nlines, column: 1000000});
+            }
+            else {
+                this.editor.setPosition({lineNumber: 1, column: 1000000});
+            }
         }
     }
 
